﻿##### Scriptinfo ##################################################################################
# Scriptreihe	    gMSA-Admin
# Datum:            2017-11-03
# Version:          V1.06
# Programmierer:    Stephan Walther
###################################################################################################

[cmdletbinding()]
param(            
)

# Variablen, Typen und Bibliotheken ###############################################################
cls

$CurrentPath = [System.IO.Path]::GetDirectoryName($myInvocation.MyCommand.Definition)

[void] [System.Reflection.Assembly]::LoadWithPartialName("System.Windows.Forms") 
[void] [System.Reflection.Assembly]::LoadWithPartialName("System.Drawing") 
[System.Reflection.Assembly]::LoadWithPartialName('Microsoft.VisualBasic') | Out-Null
[System.Windows.Forms.Application]::EnableVisualStyles()

New-Variable -Name WinRMBlackList -Value @() -Scope global
New-Variable -Name WinRMWhiteList -Value @() -Scope global

Add-Type -Name Window -Namespace Console -MemberDefinition '
[DllImport("Kernel32.dll")]
public static extern IntPtr GetConsoleWindow();
 
[DllImport("user32.dll")]
public static extern bool ShowWindow(IntPtr hWnd, Int32 nCmdShow);
'

# administrative PS ###############################################################################
$CurrentUser = [Security.Principal.WindowsPrincipal] [Security.Principal.WindowsIdentity]::GetCurrent()
If (-not ($CurrentUser).IsInRole([Security.Principal.WindowsBuiltInRole] "Administrator")) {
    Write-Host "starte administrative PowerShell ..." -ForegroundColor Cyan
    $AdminPS = new-object System.Diagnostics.ProcessStartInfo "PowerShell"
    $AdminPS.Arguments = $myInvocation.MyCommand.Definition
    $AdminPS.Verb = "runas"
    $adminps.WorkingDirectory = $CurrentPath
    [System.Diagnostics.Process]::Start($AdminPS)    
    exit 
} 

# Funktionsdeklarationen ##########################################################################
# Main
function Main {
    $PSScriptRoot = $MyInvocation.PSScriptRoot
    Set-Location $PSScriptRoot
                 
    Hide-Console |Out-Null              
    erzeuge-Hauptformular        
    Show-Console | Out-Null
}

function get-gMSA {
    Return ($obj_Liste_gMSA.SelectedItem -split ' ')[0]
}

function get-Server {
    $Server = ($obj_Liste_Server.SelectedItem -split ' ')[0]

    if ($Server -like 'S-1-5-21*') {
        Return '<SID>'
    } else {
        Return $Server
    }
}

function get-Task {
    Return (($obj_Liste_Tasks.SelectedItem -split ' ')[0] -split '\\')[1]
}

function get-gMSAService {
    Return ((($obj_Liste_Services.SelectedItem -split ' ')[0] -split '\\')[1] -split ' ')[0]
}

function prufe-WinRMStatus {
    if ($WinRMWhiteList -notcontains (get-server) -and $WinRMBLackList -notcontains (get-server)) {
        $form_Main.cursor = [System.Windows.Forms.Cursors]::WaitCursor
        $lbl_Footer.Text = "pruefe Verbindung zu Server ($(get-Server)))"
                
        $Error.Clear()
        Invoke-Command -ComputerName (get-Server) -ScriptBlock {'Test'} -ErrorAction SilentlyContinue
        if ($error) {
            $global:WinRMBlackList += (get-server)

            $Fehler = $error[0].exception.message
            [System.Windows.Forms.MessageBox]::Show("Der Computer $(get-server) konnte nicht mit WinRM ausgelesen werden: $Fehler" , "Es ist ein Fehler aufgetreten")             
        } else {
            $global:WinRMWhiteList += (get-server)
        }

        $form_Main.cursor = [System.Windows.Forms.Cursors]::Default
    }
}

function aktualisiere-TaskListe {
    param([string[]] $ServerListe, [switch]$ClearOnly=$false) 

    if ($ServerListe -eq '<SID>') {$ServerListe = @();$ClearOnly=$true}

    $form_Main.cursor = [System.Windows.Forms.Cursors]::WaitCursor
    $lbl_Footer.Text = "lade Daten von RemoteServern ($($ServerListe -join ','))"

    $Tasks = @()

    if ($ClearOnly -eq $false -and $WinRMBlackList -notcontains (get-server)) {        
        $error.Clear()
        $Tasks += Invoke-Command -ComputerName $ServerListe -ScriptBlock {
            $TaskBlackList = @()
            $TaskBlackList += 'Optimize Start Menu Cache*'

            $Tasks = Get-ScheduledTask -TaskPath '\' -ErrorAction SilentlyContinue | Where-Object {$_.state -ne 'disabled'} 
    
            $TaskBlackList | ForEach-Object {   
                $FilterWord = $_

                $cmd = "`$Tasks = `$Tasks | where-object {`$_.TaskName -notlike '$FilterWord'}"
                Invoke-Expression $cmd        
            }
        
            $Tasks | ForEach-Object {
                $env:COMPUTERNAME + '\' + $_.taskname + ' (' + $_.Principal.UserId + ')'
            }        
        } -ErrorAction SilentlyContinue
        if ($error) {
            $Fehler = $error[0].exception.message
            [System.Windows.Forms.MessageBox]::Show("Der Computer $ServerListe konnte nicht mit WinRM ausgelesen werden: $Fehler" , "Es ist ein Fehler aufgetreten")             
        }
    }
        
    $obj_Liste_Tasks.items.Clear()
    $Tasks | Sort-Object | ForEach-Object { $obj_Liste_Tasks.Items.Add($_) | Out-Null }

    $form_Main.cursor = [System.Windows.Forms.Cursors]::Default        
    $lbl_Footer.Text = 'bereit'
}

function aktualisiere-ServiceListe {
    param([string[]] $ServerListe, [switch]$ClearOnly=$false) 

    if ($ServerListe -eq '<SID>') {$ServerListe = @();$ClearOnly=$true}

    $form_Main.cursor = [System.Windows.Forms.Cursors]::WaitCursor
    $lbl_Footer.Text = "lade Daten von RemoteServern ($($ServerListe -join ','))"

    $Services = @()

    $obj_Liste_Services.items.Clear()

    if ($ClearOnly -eq $false -and $WinRMBlackList -notcontains (get-server)) {        
        $error.Clear()
        $Services = @()     
        $Services += Invoke-Command -ComputerName $ServerListe -ScriptBlock {
            param ($gMSA)
            Get-WmiObject -Query "SELECT * FROM win32_service WHERE NOT PathName LIKE '%svchost%' AND NOT StartMode = 'Disabled'" | 
                Where-Object {$_.StartName -like "*$gMSA*"} | 
                    sort-object -Property Name |
                        ForEach-Object {
                            $env:COMPUTERNAME + '\' + $_.name + ' - ' + ($_.pathname -replace '"') + " (" + $_.startname + ')'
                        } 
        } -ArgumentList (get-gMSA) -ErrorAction SilentlyContinue      
        $Services | ForEach-Object { $obj_Liste_Services.Items.Add($_) | Out-Null }

        $obj_Liste_Services.Items.Add('----------------------------------') | Out-Null

        $Services = @()
        $Services += Invoke-Command -ComputerName $ServerListe -ScriptBlock {
            param ($gMSA)
            Get-WmiObject -Query "SELECT * FROM win32_service WHERE NOT PathName LIKE '%svchost%' AND NOT StartMode = 'Disabled'" | 
                Where-Object {$_.StartName -notlike "*$gMSA*"} | 
                    sort-object -Property Name |
                        ForEach-Object {
                            $env:COMPUTERNAME + '\' + $_.name + ' - ' + ($_.pathname -replace '"') + " (" + $_.startname + ')'
                        } 
        } -ArgumentList (get-gMSA) -ErrorAction SilentlyContinue             
        $Services | ForEach-Object { $obj_Liste_Services.Items.Add($_) | Out-Null }

        if ($error) {
            $Fehler = $error[0].exception.message
            [System.Windows.Forms.MessageBox]::Show("Der Computer $ServerListe konnte nicht mit WinRM ausgelesen werden: $Fehler" , "Es ist ein Fehler aufgetreten")             
        }
    }    

    $form_Main.cursor = [System.Windows.Forms.Cursors]::Default        
    $lbl_Footer.Text = 'bereit'
}

function aktualisiere-GruppenListe {
    param([switch]$ClearOnly=$false) 

    $obj_Liste_Gruppen.Items.Clear()

    if ($ClearOnly -eq $false ) {
        $obj_Liste_Gruppen.Items.Add('--- direkte Gruppen ---') | Out-Null
        ((Get-ADServiceAccount -Identity "$(get-gMSA)`$" -Properties memberof).memberof | Get-ADGroup).name | ForEach-Object {
            $obj_Liste_Gruppen.Items.Add("   $_") | Out-Null
        }

        $obj_Liste_Gruppen.Items.Add('') | Out-Null
        $obj_Liste_Gruppen.Items.Add('--- indirekte Gruppen ---') | Out-Null
        (Get-ADAccountAuthorizationGroup -Identity "$(get-gMSA)`$").name | ForEach-Object {
            $obj_Liste_Gruppen.Items.Add("   $_") | Out-Null
        }
    }
}

function aktualisiere-ServerListe {
    param([switch]$ClearOnly=$false) 
    $obj_Liste_Server.Items.Clear()

    if ($ClearOnly -eq $false ) {
        (Get-ADServiceAccount -Filter "name -eq '$(get-gMSA)'" -Properties PrincipalsAllowedToRetrieveManagedPassword).PrincipalsAllowedToRetrieveManagedPassword |
            ForEach-Object {
                $Server = ($_ -split ',')[0] -replace 'CN='
                if ((Get-ADComputer -filter "name -eq '$Server'").DNSHostname -ne $null) {
                    $Server = (Get-ADComputer -Identity $Server).DNSHostname
                } 
                $obj_Liste_Server.Items.Add($Server) | Out-Null
            }
    }
}

function aktualisiere-gMSAListe {
    $obj_Liste_gMSA.items.Clear()
    get-ADServiceAccount -Filter * -Properties description | 
        Sort-Object -Property Samaccountname |
            ForEach-Object { 
                $gMSA_Name = $_.Name
                $gMSA_Desc = $_.description
                if ($gMSA_desc.length -gt 0) {$gMSA_desc = ' (' + $gMSA_desc + ')'}
                $obj_Liste_gMSA.Items.Add($gMSA_Name + $gMSA_desc) | Out-Null 
            }
}
  
# Formular
function erzeuge-Hauptformular {
    $form_Main = New-Object System.Windows.Forms.Form 
    $form_Main.Text = "gMSA-Admin"
    $form_Main.Size = '1024,800'
    $form_Main.StartPosition = "CenterScreen"
    $form_Main.Topmost = $false
    $form_Main.KeyPreview = $True
    $Logo = (Get-Item -Path .\wsits.ico -ErrorAction SilentlyContinue).FullName
    if ($Logo) { $form_Main.Icon = $Logo }
    
    $lbl_Liste_gMSA = New-Object System.Windows.Forms.Label
    $lbl_Liste_gMSA.Location = '10,10'
    $lbl_Liste_gMSA.Size = '200,20'
    $lbl_Liste_gMSA.Text = "vorhandene gMSA:"
    $form_Main.Controls.Add($lbl_Liste_gMSA)    
        
    $obj_Liste_gMSA = New-Object System.Windows.Forms.ListBox
    $obj_Liste_gMSA.Location = '10,35'
    $obj_Liste_gMSA.Size = '300,300'
    $obj_Liste_gMSA.add_SelectedIndexChanged({ 
        # aktualisiere Listen             
            aktualisiere-ServerListe
            aktualisiere-GruppenListe
            aktualisiere-TaskListe -ClearOnly
            aktualisiere-ServiceListe -ClearOnly

        # setze Schalter            
            $Button_loescheGMSA.enabled = $true
            $Button_neuerServer.enabled = $true
            $Button_neueGruppe.enabled = $true 
            $Button_TaskLeseAlle.Enabled = $true   
            $Button_TaskSetzeGMSA.Enabled = $false    
            $Button_ServiceSetzeGMSA.enabled = $false   
            $Button_entferneServer.enabled = $false
            $Button_ServerTesteGMSA.enabled = $false    
            $Button_ServiceReset.enabled = $false      
    })
    $form_Main.Controls.Add($obj_Liste_gMSA) 
    
    $Button_erstelleGMSA = New-Object System.Windows.Forms.Button
    $Button_erstelleGMSA.Location = '10,335'
    $Button_erstelleGMSA.Size = '100,25'
    $Button_erstelleGMSA.Text = "erstelle gMSA"                                                                                                                        
    $Button_erstelleGMSA.Add_Click({ 
        $neuerGMSA = [Microsoft.VisualBasic.Interaction]::InputBox("Geben Sie den Namen des gMSA ein", "neuer gMSA", "")    
        $neue_Desc = [Microsoft.VisualBasic.Interaction]::InputBox("Geben Sie eine Beschreibung ein", "neuer gMSA", "")    
        if ($neuerGMSA.length -gt 0) {
            # erstelle gMSA
                New-ADServiceAccount `
                    -Name $neuerGMSA `
                    -DNSHOSTNAME "$neuerGMSA.$Domain" `
                    -Description $neue_Desc

            # aktualisiere Listen
                aktualisiere-gMSAListe                
                $obj_Liste_gMSA.SetSelected($obj_Liste_gMSA.FindString($neuerGMSA),$true) 
            
                aktualisiere-ServerListe -ClearOnly
                aktualisiere-GruppenListe
                aktualisiere-TaskListe -ClearOnly
                aktualisiere-ServiceListe -ClearOnly

            # setze Schalter            
                $Button_loescheGMSA.enabled = $true
                $Button_neuerServer.enabled = $true
                $Button_neueGruppe.enabled = $true 
                $Button_TaskLeseAlle.Enabled = $true   
                $Button_TaskSetzeGMSA.Enabled = $false    
                $Button_ServiceSetzeGMSA.enabled = $false
                $Button_ServiceReset.enabled = $false
        }
    })    
    $form_Main.Controls.Add($Button_erstelleGMSA)

    $Button_loescheGMSA = New-Object System.Windows.Forms.Button
    $Button_loescheGMSA.Location = '110,335'
    $Button_loescheGMSA.Size = '100,25'
    $Button_loescheGMSA.Text = "lösche gMSA"                                                                                                                        
    $Button_loescheGMSA.Enabled = $false
    $Button_loescheGMSA.Add_Click({            
        if ([System.Windows.Forms.MessageBox]::Show("Soll der gMSA geloescht werden?" , "Warnung",[System.Windows.Forms.MessageBoxButtons]::YesNo) -eq 'YES') {
            Remove-ADServiceAccount -Identity (get-gMSA) -Confirm:$false

            aktualisiere-gMSAListe
            aktualisiere-ServerListe -ClearOnly
            aktualisiere-GruppenListe -ClearOnly
            aktualisiere-TaskListe -ClearOnly
            aktualisiere-ServiceListe -ClearOnly

            $Button_entferneServer.enabled = $false
            $Button_neuerServer.enabled = $false
            $Button_ServerTesteGMSA.Enabled = $false
            $Button_loescheGMSA.Enabled = $false  
            $Button_neueGruppe.Enabled = $false
            $Button_entferneGruppe.Enabled = $false
            $Button_TaskLeseAlle.Enabled = $false
            $Button_TaskSetzeGMSA.Enabled = $false
            $Button_ServiceSetzeGMSA.enabled = $false
            $Button_ServiceReset.enabled = $false
        }
    })    
    $form_Main.Controls.Add($Button_loescheGMSA)

    $lbl_Liste_Server = New-Object System.Windows.Forms.Label
    $lbl_Liste_Server.Location = '320,10'
    $lbl_Liste_Server.Size = '200,20'
    $lbl_Liste_Server.Text = "zugehörige Server:"
    $form_Main.Controls.Add($lbl_Liste_Server) 
        
    $obj_Liste_Server = New-Object System.Windows.Forms.ListBox
    $obj_Liste_Server.Location = '320,35'
    $obj_Liste_Server.Size = '300,300'
    $obj_Liste_Server.add_SelectedIndexChanged({ 
        # Schalter aktivieren
            $Button_entferneServer.enabled = $true  
            $Button_ServerTesteGMSA.Enabled = $true  
        
        # Listen aktualisieren   
            prufe-WinRMStatus         
            aktualisiere-TaskListe -ServerListe (get-Server)
            aktualisiere-ServiceListe -ServerListe (get-server)

        # Status aktualisieren
            if ($WinRMWhiteList -contains (get-server)) { $lbl_Footer.Text = "bereit" }
            if ($WinRMBlackList -contains (get-server)) { $lbl_Footer.Text = "keine Verbindung zu Server ($(get-Server)))" }
    })
    $form_Main.Controls.Add($obj_Liste_Server) 

    $Button_neuerServer = New-Object System.Windows.Forms.Button
    $Button_neuerServer.Location = '320,335'
    $Button_neuerServer.Size = '100,25'
    $Button_neuerServer.Text = "weiterer Server" 
    $Button_neuerServer.Enabled = $false                                                                                                                       
    $Button_neuerServer.Add_Click({ 
        $neuerServer = [Microsoft.VisualBasic.Interaction]::InputBox("Geben Sie den Namen des Servers ein", "neuer Server für gMSA", "") 

        if ($neuerServer.length -gt 0) {                        
            if (-not (Get-ADComputer -Filter "name -eq '$neuerServer'")) {
                [System.Windows.Forms.MessageBox]::Show("Der Computer $neuerServer wurde nicht gefunden!" , "Es ist ein Fehler aufgetreten") 
            } else {
                # ermittle aktuelle Server       
                    $MemberServer =  @()    
                    $MemberServer += (Get-ADServiceAccount -Filter "name -eq '$(get-gMSA)'" -Properties PrincipalsAllowedToRetrieveManagedPassword).PrincipalsAllowedToRetrieveManagedPassword
                    $MemberServer += Get-ADComputer -Filter "name -eq '$neuerServer'"

                # füge neuen Server an
                    $error.Clear()
                    Set-ADServiceAccount `
                        -identity (get-gMSA) `
                        -PrincipalsAllowedToRetrieveManagedPassword $MemberServer `
                        -ErrorAction SilentlyContinue
                    if ($error) {
                        $Fehler = $error[0].Exception.Message
                        [System.Windows.Forms.MessageBox]::Show("$Fehler" , "Es ist ein Fehler aufgetreten") 
                    } else {
                        aktualisiere-ServerListe
                        #aktualisiere-TaskListe -ServerListe $neuerServer
                        #aktualisiere-ServiceListe -ServerListe $neuerServer

                        $obj_Liste_Server.SetSelected($obj_Liste_Server.FindString($neuerServer),$true) 

                        $Button_entferneServer.enabled = $true
                        $Button_ServerTesteGMSA.Enabled = $true
                        $Button_TaskSetzeGMSA.enabled = $false
                        $Button_ServiceSetzeGMSA.enabled = $false
                        $Button_ServiceReset.enabled = $false
                    }
            }
        }
                           
        
    })    
    $form_Main.Controls.Add($Button_neuerServer)

    $Button_entferneServer = New-Object System.Windows.Forms.Button
    $Button_entferneServer.Location = '420,335'
    $Button_entferneServer.Size = '100,25'
    $Button_entferneServer.Text = "entferne Server"      
    $Button_entferneServer.Enabled = $false                                                                                                                                                                                                                                         
    $Button_entferneServer.Add_Click({ 
        # ermittle verbleibende Server
            $MemberServer = @()
            (Get-ADServiceAccount -Filter "name -eq '$(get-gMSA)'" -Properties PrincipalsAllowedToRetrieveManagedPassword).PrincipalsAllowedToRetrieveManagedPassword |
                ForEach-Object {
                    $MemberServer += ($_ -split ',')[0] -replace 'CN='
                }
            $MemberServer = $MemberServer -notmatch $obj_Liste_Server.SelectedItem
            #$MemberServer = $MemberServer | Get-ADComputer
        
        # schreibe Serverliste zurück
            $error.Clear()
            Set-ADServiceAccount `
                -identity (get-gMSA) `
                -PrincipalsAllowedToRetrieveManagedPassword $MemberServer `
                -ErrorAction SilentlyContinue
            if ($error) {
                $Fehler = $error[0].Exception.Message
                [System.Windows.Forms.MessageBox]::Show("$Fehler" , "Es ist ein Fehler aufgetreten") 
            }
        # ermittle Server für Ausgabe
            aktualisiere-ServerListe
            aktualisiere-TaskListe -ClearOnly
            aktualisiere-ServiceListe -ClearOnly

        # deaktivere Button
            $Button_entferneServer.enabled = $false
            $Button_ServerTesteGMSA.Enabled = $false
            $Button_TaskSetzeGMSA.Enabled = $false
            $Button_ServiceSetzeGMSA.enabled = $false
            $Button_ServiceReset.enabled = $false
    })    
    $form_Main.Controls.Add($Button_entferneServer)

    $Button_ServerTesteGMSA = New-Object System.Windows.Forms.Button
    $Button_ServerTesteGMSA.Location = '520,335'
    $Button_ServerTesteGMSA.Size = '100,25'
    $Button_ServerTesteGMSA.Text = "teste gMSA"                                                                                                                        
    $Button_ServerTesteGMSA.Enabled = $false
    $Button_ServerTesteGMSA.Add_Click({        
        $form_Main.cursor = [System.Windows.Forms.Cursors]::WaitCursor
        $lbl_Footer.Text = "lade Daten von RemoteServern ($(get-Server)))"
        
        $Ergebnis = Invoke-Command -ComputerName (get-Server) -ScriptBlock {
            param($Server,$gMSA)  
            
            $error.Clear()
                        
            New-Item -Path "c:\gMSA-Test"  -ItemType directory -ErrorAction SilentlyContinue | Out-Null
            "whoami > c:\gMSA-Test\result.txt" | Out-File -FilePath "c:\gMSA-Test\script.bat" -Encoding default -ErrorAction SilentlyContinue
            
            $action    = New-ScheduledTaskAction  "c:\gMSA-Test\script.bat" 
            $trigger   = New-ScheduledTaskTrigger -AtLogOn
            $principal = New-ScheduledTaskPrincipal -UserID "$gMSA" -LogonType Password
            Register-ScheduledTask "gMSA-Test" –Action $action –Trigger $trigger –Principal $principal -ErrorAction SilentlyContinue | Out-Null

            if ($error) {
                "Fehler: $($error | foreach {$_.exception.message})"
            } else {
                Wait-Event -Timeout 1
                Start-ScheduledTask -TaskName "gMSA-Test"
                Wait-Event -Timeout 1

                if (Test-Path -Path "c:\gMSA-Test\result.txt") {
                    $user = Get-Content -Path "c:\gMSA-Test\result.txt"
                    if ($user -eq $gMSA ) {
                        "OK"
                    } else {
                        "Fehler ($user <> $gMSA)"
                    }
                } else {
                    "Fehler: $($error | foreach {$_.exception.message})"
                }

                Remove-Item -Path "c:\gMSA-Test" -Recurse -ErrorAction SilentlyContinue

                Unregister-ScheduledTask "gMSA-Test" -Confirm:$false -ErrorAction SilentlyContinue
            }

        } -ArgumentList (get-Server),"$($env:USERDOMAIN)\$(get-gMSA)$" 
        
        $form_Main.cursor = [System.Windows.Forms.Cursors]::Default        
        $lbl_Footer.Text = 'bereit' 

        if ($Ergebnis -eq 'OK') {               
            [System.Windows.Forms.MessageBox]::Show("Der ServiceAccount kann als Task verwendet werden!" , "Erfolg") 
        } else {
            [System.Windows.Forms.MessageBox]::Show("Der ServiceAccount kann nicht als Task verwendet werden: $Ergebnis" , "Es ist ein Fehler aufgetreten") 
        }           
    })    
    $form_Main.Controls.Add($Button_ServerTesteGMSA)
    
    $lbl_Liste_Gruppen = New-Object System.Windows.Forms.Label
    $lbl_Liste_Gruppen.Location = '630,10'
    $lbl_Liste_Gruppen.Size = '200,20'
    $lbl_Liste_Gruppen.Text = "zugehörige Gruppen:"
    $form_Main.Controls.Add($lbl_Liste_Gruppen) 
        
    $obj_Liste_Gruppen = New-Object System.Windows.Forms.ListBox
    $obj_Liste_Gruppen.Location = '630,35'
    $obj_Liste_Gruppen.Size = '300,300'
    $obj_Liste_Gruppen.add_SelectedIndexChanged({ 
          $Button_entferneGruppe.enabled = $true  
    })
    $form_Main.Controls.Add($obj_Liste_Gruppen) 

    $Button_neueGruppe = New-Object System.Windows.Forms.Button
    $Button_neueGruppe.Location = '630,335'
    $Button_neueGruppe.Size = '100,25'
    $Button_neueGruppe.Text = "weitere Gruppe" 
    $Button_neueGruppe.Enabled = $false                                                                                                                       
    $Button_neueGruppe.Add_Click({ 
        $neueGruppe = [Microsoft.VisualBasic.Interaction]::InputBox("Geben Sie den Namen der Gruppe ein", "gMSA in Gruppe aufnehmen", "") 
        if ($neueGruppe.length -gt 0) {                        
            if (-not (Get-ADGroup -Filter "name -eq '$neueGruppe'" -ErrorAction SilentlyContinue)) {
                [System.Windows.Forms.MessageBox]::Show("Die Gruppe $neueGruppe wurde nicht gefunden!" , "Es ist ein Fehler aufgetreten") 
            } else {
                # Aufnahme in Gruppe
                    Add-ADGroupMember -Identity $neueGruppe -Members "$(get-gMSA)`$"
                
                # ermittle Gruppen für Ausgabe
                    aktualisiere-GruppenListe
                    $obj_Liste_Gruppen.SetSelected($obj_Liste_Gruppen.FindString("   $neueGruppe"),$true) 

                    $Button_entferneGruppe.enabled = $true
            }
        }
    })    
    $form_Main.Controls.Add($Button_neueGruppe)

    $Button_entferneGruppe = New-Object System.Windows.Forms.Button
    $Button_entferneGruppe.Location = '730,335'
    $Button_entferneGruppe.Size = '100,25'
    $Button_entferneGruppe.Text = "entferne Gruppe"      
    $Button_entferneGruppe.Enabled = $false                                                                                                                                                                                                                                         
    $Button_entferneGruppe.Add_Click({ 
        # entferne Mitgliedschaft
            Remove-ADGroupMember -Identity "$($obj_Liste_Gruppen.SelectedItem)" -Members "$(get-gMSA)`$" -Confirm:$false
            
        # ermittle Gruppen für Ausgabe
            aktualisiere-GruppenListe

        # deaktivere Button
            $Button_entferneGruppe.enabled = $false
    })    
    $form_Main.Controls.Add($Button_entferneGruppe)
    
    $lbl_Liste_Tasks = New-Object System.Windows.Forms.Label
    $lbl_Liste_Tasks.Location = '10,375'
    $lbl_Liste_Tasks.Size = '200,20'
    $lbl_Liste_Tasks.Text = "vorhandene Tasks:"
    $form_Main.Controls.Add($lbl_Liste_Tasks)    
        
    $obj_Liste_Tasks = New-Object System.Windows.Forms.ListBox
    $obj_Liste_Tasks.Location = '10,400'
    $obj_Liste_Tasks.Size = '300,300'
    $obj_Liste_Tasks.add_SelectedIndexChanged({ 
        $Button_TaskSetzeGMSA.Enabled = $true
    })
    $form_Main.Controls.Add($obj_Liste_Tasks) 

    $Button_TaskLeseAlle = New-Object System.Windows.Forms.Button
    $Button_TaskLeseAlle.Location = '10,700'
    $Button_TaskLeseAlle.Size = '100,25'
    $Button_TaskLeseAlle.Text = "lese alle Server"  
    $Button_TaskLeseAlle.Enabled = $false                                                                                                                      
    $Button_TaskLeseAlle.Add_Click({ 
        $ServerListe = @()
        (Get-ADServiceAccount -Filter "name -eq '$(get-gMSA)'" -Properties PrincipalsAllowedToRetrieveManagedPassword).PrincipalsAllowedToRetrieveManagedPassword |
            ForEach-Object {
                $ServerListe += ($_ -split ',')[0] -replace 'CN='
            }
        aktualisiere-TaskListe -ServerListe $ServerListe
    })    
    $form_Main.Controls.Add($Button_TaskLeseAlle)

    $Button_TaskSetzeGMSA = New-Object System.Windows.Forms.Button
    $Button_TaskSetzeGMSA.Location = '110,700'
    $Button_TaskSetzeGMSA.Size = '100,25'
    $Button_TaskSetzeGMSA.Text = "setze gMSA ein"                                                                                                                        
    $Button_TaskSetzeGMSA.Enabled = $false
    $Button_TaskSetzeGMSA.Add_Click({          
        $Ergebnis = Invoke-Command -ComputerName (get-Server) -ScriptBlock {
            param($TaskName,$gMSA)  
        
            $principal = (Get-ScheduledTask -TaskName $TaskName).principal
            $principal.UserId = $gMSA
    
            Set-ScheduledTask -TaskName $TaskName -Principal $principal -ErrorAction SilentlyContinue

            if ($error) {"Fehler: $($error | foreach {$_.exception.message})"} else {"OK"}
        } -ArgumentList (get-Task),"$($env:USERDOMAIN)\$(get-gMSA)$" 
        
        if ($Ergebnis -eq 'OK') {
            aktualisiere-TaskListe -ServerListe (get-server)                    
            $obj_Liste_Tasks.SetSelected($obj_Liste_Tasks.FindString((get-Task)),$true) 
            
            [System.Windows.Forms.MessageBox]::Show("Der Task wurde umgestellt!" , "Erfolg") 
        } else {
            [System.Windows.Forms.MessageBox]::Show("Der Task konnte nicht umgestellt werden: $Ergebnis" , "Es ist ein Fehler aufgetreten") 
        }
        
    })    
    $form_Main.Controls.Add($Button_TaskSetzeGMSA)

    $lbl_Liste_Services= New-Object System.Windows.Forms.Label
    $lbl_Liste_Services.Location = '320,375'
    $lbl_Liste_Services.Size = '200,20'
    $lbl_Liste_Services.Text = "vorhandene Services:"
    $form_Main.Controls.Add($lbl_Liste_Services)    
        
    $obj_Liste_Services = New-Object System.Windows.Forms.ListBox
    $obj_Liste_Services.Location = '320,400'
    $obj_Liste_Services.Size = '610,300'
    $obj_Liste_Services.add_SelectedIndexChanged({ 
        $Button_ServiceSetzeGMSA.Enabled = $true
        $Button_ServiceReset.enabled = $true
    })
    $form_Main.Controls.Add($obj_Liste_Services) 

    <#$obj_Liste_Services = New-Object System.Windows.Forms.DataGridView  
    $obj_Liste_Services.Location = '320,400'
    $obj_Liste_Services.Size='610,300'
    $obj_Liste_Services.RowHeadersVisible = $true
    $obj_Liste_Services.AllowUserToAddRows = $false
    $obj_Liste_Services.AllowUserToDeleteRows = $false 
    $obj_Liste_Services.Rows.Clear()
    $obj_Liste_Services.ColumnCount = 4     
    $obj_Liste_Services.ColumnHeadersVisible = $true
    $obj_Liste_Services.Columns[0].Name = "Server"
    $obj_Liste_Services.Columns[0].Width = 130
    $obj_Liste_Services.Columns[1].Name = "Service"
    $obj_Liste_Services.Columns[1].Width = 130
    $obj_Liste_Services.Columns[2].Name = "Account"
    $obj_Liste_Services.Columns[2].Width = 75
    $obj_Liste_Services.Columns[3].Name = "Pfad"
    $obj_Liste_Services.Columns[3].Width = 300
    $obj_Liste_Services.Rows.Add('Zelle1','Zelle2','Zelle3') | Out-Null
    $form_Main.Controls.Add($obj_Liste_Services) 
    #>

    $Button_ServicesLeseAlle = New-Object System.Windows.Forms.Button
    $Button_ServicesLeseAlle.Location = '320,700'
    $Button_ServicesLeseAlle.Size = '100,25'
    $Button_ServicesLeseAlle.Text = "lese alle Server"  
    $Button_ServicesLeseAlle.Enabled = $false                                                                                                                      
    $Button_ServicesLeseAlle.Add_Click({ 
       
    })    
    $form_Main.Controls.Add($Button_ServicesLeseAlle)

    $Button_ServiceSetzeGMSA = New-Object System.Windows.Forms.Button
    $Button_ServiceSetzeGMSA.Location = '420,700'
    $Button_ServiceSetzeGMSA.Size = '100,25'
    $Button_ServiceSetzeGMSA.Text = "setze gMSA ein"                                                                                                                        
    $Button_ServiceSetzeGMSA.Enabled = $false
    $Button_ServiceSetzeGMSA.Add_Click({          
        $Ergebnis = Invoke-Command -ComputerName (get-Server) -ScriptBlock {        
            param ($ServiceName,$gMSA)

            $gMSA = $env:USERdomain + "\$gMSA`$"

            $Service = Get-WmiObject -Query "SELECT * FROM win32_service WHERE name = '$ServiceName'"
            $Ergebnis = $Service.Change($null,$null,$null,$null,$null,$null,$gMSA,$Null,$null,$null,$null)

            if ($Ergebnis.ReturnValue -eq 0) {
                Restart-Service $ServiceName -ErrorAction SilentlyContinue -WarningAction SilentlyContinue
            
                if ($Error){"Fehler: $($error | foreach {$_.exception.message})"} else {'OK'}                      
            } else {
                "Service.Change == $Ergebnis.ReturnValue"
            }
            
        } -ArgumentList (get-gMSAService),(get-gMSA)

        if ($Ergebnis -eq 'OK') {
            aktualisiere-ServiceListe -ServerListe (get-Server)
            $obj_Liste_Services.SetSelected($obj_Liste_Services.FindString((get-gMSAService)),$true) 
            
            [System.Windows.Forms.MessageBox]::Show("Der Service wurde neu gestartet!" , "Erfolg") 
        } else {
            [System.Windows.Forms.MessageBox]::Show("Der Service konnte nicht neu gestartet werden: $Ergebnis!" , "Es ist ein Fehler aufgetreten") 
        }        
    })    
    $form_Main.Controls.Add($Button_ServiceSetzeGMSA)

    $Button_ServiceReset = New-Object System.Windows.Forms.Button
    $Button_ServiceReset.Location = '520,700'
    $Button_ServiceReset.Size = '100,25'
    $Button_ServiceReset.Text = "zurücksetzen"                                                                                                                        
    $Button_ServiceReset.Enabled = $false
    $Button_ServiceReset.Add_Click({          
        $Ergebnis = Invoke-Command -ComputerName (get-Server) -ScriptBlock {        
            param ($ServiceName)
            
            $Service = Get-WmiObject -Query "SELECT * FROM win32_service WHERE name = '$ServiceName'"
            $Ergebnis = $Service.Change($null,$null,$null,$null,$null,$null,'localsystem',$Null,$null,$null,$null)

            if ($Ergebnis.ReturnValue -eq 0) {
                Restart-Service $ServiceName -ErrorAction SilentlyContinue -WarningAction SilentlyContinue
            
                if ($Error){"Fehler: $($error | foreach {$_.exception.message})"} else {'OK'}                      
            } else {
                "Service.Change == $Ergebnis.ReturnValue"
            }
            
        } -ArgumentList (get-gMSAService)

        if ($Ergebnis -eq 'OK') {
            aktualisiere-ServiceListe -ServerListe (get-Server)
                        
            [System.Windows.Forms.MessageBox]::Show("Der Service wurde neu gestartet!" , "Erfolg") 
        } else {
            [System.Windows.Forms.MessageBox]::Show("Der Service konnte nicht neu gestartet werden: $Ergebnis!" , "Es ist ein Fehler aufgetreten") 
        }        
    })    
    $form_Main.Controls.Add($Button_ServiceReset)
            
    $lbl_Footer = New-Object 'System.Windows.Forms.Label'
    $lbl_Footer.Location = "10,$($form_Main.Height - 60)"
    $lbl_Footer.Size = "$($form_Main.Height - 30),20"
    $lbl_Footer.text = 'Bereit'    
    $form_Main.Controls.Add($lbl_Footer)
           
    $form_Main.ResumeLayout()	
    $InitialFormWindowState = $form_Main.WindowState
    $form_Main.Add_Shown({
        $form_Main.Activate()
        $form_Main.cursor = [System.Windows.Forms.Cursors]::WaitCursor
        $lbl_Footer.Text = 'lade Daten'
        
        Import-Module activedirectory

        $StopScript = $false

        if ((Get-KdsRootKey) -eq $null) { 
            $Antwort = [System.Windows.Forms.MessageBox]::Show(
                "Im AD wurde kein KDSRootKey gefunden. Soll ein Key erstellt werden?", 
                "Prüfung der gMSA-Voraussetzungen",
                [System.Windows.Forms.MessageBoxButtons]::YesNo
            ) 
            if ($Antwort -eq 'yes') { 
                $Version = (Get-ADDomain).DomainMode
                if (2000,2003,2008 -contains ($Version -replace '\D').substring(0,4)) {              
                    $Antwort = [System.Windows.Forms.MessageBox]::Show(
                        "Fuer gMSA muss die Funktionsebene der Domain mindestens 2012 sein. Die aktuelle Version ist '$Version'. Soll die Funktionsebene hochgesetzt werden? Der Prozess kann nicht zurueckgenommen werden!!!", 
                        "Prüfung der gMSA-Voraussetzungen",
                        [System.Windows.Forms.MessageBoxButtons]::YesNo
                    )   
                    if ($Antwort -eq 'yes') { 
                        $cred = (Get-Credential -Message 'Geben Sie Anmeldeinformationen von einem Benutzer ein, der die Funktionsebene aendern darf')
                        $error.Clear()
                        if ($cred) { 
                            Show-Console
                            Get-ADDomain | Set-ADDomainMode -DomainMode Windows2012Domain -Credential $cred 
                            Hide-Console
                        } else { $StopScript = $true }
                        if ($error) { 
                            [System.Windows.Forms.MessageBox]::Show(
                                "Es ist ein Fehler aufgetreten. Die Funktionsebene wurde nicht geaendert. Es kann kein Key erstellt werden. Das Script wird nun beendet!", 
                                "Prüfung der gMSA-Voraussetzungen"
                            )
                            $StopScript = $true 
                        } else {
                            [System.Windows.Forms.MessageBox]::Show(
                                "Die Funktionsebene des AD wurde auf Windows2012Domain aktualisiert. Ggf. muss eine AD-Replikation abgewartet werden!",
                                "Prüfung der gMSA-Voraussetzungen"
                            )
                        }
                    } else {
                        $StopScript = $true 
                    }
                }    
                if ( $StopScript -eq $false ) {
                    $error.Clear()
                    if ($cred -eq $null) { $cred = (Get-Credential -Message 'Geben Sie Anmeldeinformationen von einem Benutzer ein, der einen KDSRootKey erstellen darf') }
                    if ($cred) { Add-KdsRootKey -EffectiveTime (Get-Date).AddHours(-10) } else { $StopScript = $true }
                    
                    if ($error) { 
                        [System.Windows.Forms.MessageBox]::Show(
                            "Es ist ein Fehler aufgetreten. Es konnte kein Key erstellt werden!", 
                            "Prüfung der gMSA-Voraussetzungen"
                        )
                        $StopScript = $true 
                    }
                }            
            } else {
                $StopScript = $true 
            }
        }
        if ( $StopScript -eq $true ) {
            $Button_erstelleGMSA.Enabled = $false
            $lbl_Footer.Text = 'Es wurde kein KDS-RootKey gefunden!'            
        } else {
            aktualisiere-gMSAListe
            $lbl_Footer.Text = 'bereit'
        }

        $form_Main.cursor = [System.Windows.Forms.Cursors]::Default        
        
    })    	
	$form_Main.add_Load({
        
    })
    [void] $form_Main.ShowDialog()    
}

# zeige CLI
function Show-Console {
    $consolePtr = [Console.Window]::GetConsoleWindow()
    [Console.Window]::ShowWindow($consolePtr, 5)
}
 
# verstecke CLI
function Hide-Console {
    $consolePtr = [Console.Window]::GetConsoleWindow()
    [Console.Window]::ShowWindow($consolePtr, 0)
}

# Programmstart ###################################################################################
MAIN

# SIG # Begin signature block
# MIII4QYJKoZIhvcNAQcCoIII0jCCCM4CAQExCzAJBgUrDgMCGgUAMGkGCisGAQQB
# gjcCAQSgWzBZMDQGCisGAQQBgjcCAR4wJgIDAQAABBAfzDtgWUsITrck0sYpfvNR
# AgEAAgEAAgEAAgEAAgEAMCEwCQYFKw4DAhoFAAQUb4qwM9onJC9VkcYSrtdNjNO4
# 4Y+gggY/MIIGOzCCBSOgAwIBAgITGwAAAOBAvnzebcAjSAABAAAA4DANBgkqhkiG
# 9w0BAQsFADBUMRMwEQYKCZImiZPyLGQBGRYDaXRzMRIwEAYKCZImiZPyLGQBGRYC
# d3MxKTAnBgNVBAMTIFdTLUlUUy1aZXJ0aWZpemllcnVuZ3NzdGVsbGUtQ0ExMB4X
# DTE3MDUxMDA1NDI0OVoXDTE5MDUxMDA1NTI0OVowdTETMBEGCgmSJomT8ixkARkW
# A2l0czESMBAGCgmSJomT8ixkARkWAndzMQswCQYDVQQLEwJXUzERMA8GA1UECxMI
# QmVudXR6ZXIxDzANBgNVBAsTBldTLUlUUzEZMBcGA1UEAxMQV2FsdGhlciwgU3Rl
# cGhhbjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCCAQoCggEBAKqnKBq0nJLvqGwO
# aJkLU2174/KhLd4ykJZ2BgG4wse7s3ALM7dH1vxNmeKIoB09yk9nYZzDeRjZ7w7a
# WKgko8zvNHJCleQNtG6AReOoMIRuYyrbUmbKlmICn9V/ENUNQ6s1tiqnAIChb0J2
# 05GMaBYG+IFS+uLDzlTIsoaEKZjV3htgBEVQt6BYIdbsZ913ZUcpGWhQsCXKairT
# wB8pqPTaBCA5HJHyfAwxa+w2ERhF5q0r/xp9dEkemFjy0Xo9aQWUw3WPy8l3pTyp
# I2w/yDFqLfZu5z1QjWrcNA9VirSiM9PkOPXQb9riChc4PfFd90+JjeVhOlj3y0qb
# InQsfUMCAwEAAaOCAuMwggLfMA4GA1UdDwEB/wQEAwIHgDA8BgkrBgEEAYI3FQcE
# LzAtBiUrBgEEAYI3FQiIoH6B98FWhLmPAobjm2SE8KUnMYPz1QqE8ow3AgFkAgEC
# MB0GA1UdDgQWBBT1lyML2Qw04hBbAlX5PeqiA5IBQjAfBgNVHSMEGDAWgBS1Nrmv
# a6rveo5pQqKi2+u+kQ59cTCCASAGA1UdHwSCARcwggETMIIBD6CCAQugggEHhoHE
# bGRhcDovLy9DTj1XUy1JVFMtWmVydGlmaXppZXJ1bmdzc3RlbGxlLUNBMSgxKSxD
# Tj1XUy1DQTEsQ049Q0RQLENOPVB1YmxpYyUyMEtleSUyMFNlcnZpY2VzLENOPVNl
# cnZpY2VzLENOPUNvbmZpZ3VyYXRpb24sREM9d3MsREM9aXRzP2NlcnRpZmljYXRl
# UmV2b2NhdGlvbkxpc3Q/YmFzZT9vYmplY3RDbGFzcz1jUkxEaXN0cmlidXRpb25Q
# b2ludIY+aHR0cDovL2NybC53cy5pdHMvY3JsZC9XUy1JVFMtWmVydGlmaXppZXJ1
# bmdzc3RlbGxlLUNBMSgxKS5jcmwwgc0GCCsGAQUFBwEBBIHAMIG9MIG6BggrBgEF
# BQcwAoaBrWxkYXA6Ly8vQ049V1MtSVRTLVplcnRpZml6aWVydW5nc3N0ZWxsZS1D
# QTEsQ049QUlBLENOPVB1YmxpYyUyMEtleSUyMFNlcnZpY2VzLENOPVNlcnZpY2Vz
# LENOPUNvbmZpZ3VyYXRpb24sREM9d3MsREM9aXRzP2NBQ2VydGlmaWNhdGU/YmFz
# ZT9vYmplY3RDbGFzcz1jZXJ0aWZpY2F0aW9uQXV0aG9yaXR5MBMGA1UdJQQMMAoG
# CCsGAQUFBwMDMBsGCSsGAQQBgjcVCgQOMAwwCgYIKwYBBQUHAwMwKQYDVR0RBCIw
# IKAeBgorBgEEAYI3FAIDoBAMDnN0ZXBoYW5Ad3MuaXRzMA0GCSqGSIb3DQEBCwUA
# A4IBAQBjyTnnRXeyMEPW8SziBVFJcsX0K3XLCXz7PEl6sVuiWTHY3n9IOwJJ2sqF
# 3Sea0bpWrDFnJQtB+S4FXW6A/bVvuPlGMBtZ7T9DowowAQAmWvC/OJeeHGz2xyjh
# fDYAmMi3DXrVIEv+yWHsrh6CSt0KiHjMOh/Kk092AmyyPSRZpqmabzcoY7fI6xff
# ckOvS+Q3233ytXfS2YMuNKL8EuGaUhGj8Smb3grUtNW0kyJmL2wTgyWdUerbGpC3
# Ss7iv53hJXSoe8Rmo8spntb/5e+zt6q0ThqI4TN6sm9Gqgg8QcfzGc0XvW7V2GKU
# LT75dOMzV86lFZo+rvq1SaZb2GRcMYICDDCCAggCAQEwazBUMRMwEQYKCZImiZPy
# LGQBGRYDaXRzMRIwEAYKCZImiZPyLGQBGRYCd3MxKTAnBgNVBAMTIFdTLUlUUy1a
# ZXJ0aWZpemllcnVuZ3NzdGVsbGUtQ0ExAhMbAAAA4EC+fN5twCNIAAEAAADgMAkG
# BSsOAwIaBQCgeDAYBgorBgEEAYI3AgEMMQowCKACgAChAoAAMBkGCSqGSIb3DQEJ
# AzEMBgorBgEEAYI3AgEEMBwGCisGAQQBgjcCAQsxDjAMBgorBgEEAYI3AgEVMCMG
# CSqGSIb3DQEJBDEWBBQditWeUnAOBEVI7pobRfBfCAAnXzANBgkqhkiG9w0BAQEF
# AASCAQCIAjNgniICyn6WHxgFiDME/mnrLNMG2kOVVnQ63nYQdSoCosIwUlnkpk52
# Bi1ZF8di93IdY1FOu3Jl7wwyCtsY40NKh0lARYTD6NQGaUTmVsRUEnz4BLkXEdog
# UBLKzCTxQAeu9kJY9A43Mfk3kBLv59er+flXrDEJDdIe5BrB6ZYPuRiarhEu4J7Z
# 3MA0Hx4ik1mBIjhkKzRtM1dfHrWC9KzxjgoO4GOkkNXWpJLds5nCl43/9NzixqrL
# qP7Alw6v52wD5op6LMnMDECKhDhVZbeU0TKzjxx5L2H2F3GWYiMiA58PX/ZwsCk9
# S+nsoEoZ2K2IuMHivcF/o3Muk+2g
# SIG # End signature block
